"""
Create a class for storing a large number of simulations.

Hierarchy
- result: parameters, seed, beta
- samples: collection of results with same parameters but different seeds
- dataset: collection of samples (with different parameters)
"""

import io
import zipfile
import sciris as sc

__all__ = ['Dataset', 'Samples']

class Dataset:
    def __init__(self, folder=None, results=None, *args, **kwargs):
        # Note that results are not deep copied, to save memory
        if folder is not None:
            self.results = sc.odict()
            for file in sorted(sc.path(folder).iterdir()):
                if file.suffix == ".zip":
                    res = Samples(file, *args, **kwargs)
                    self.results[res.identifier] = res
        elif results is not None:
            self.results = results
        else:
            raise Exception("Must provide either folder or results to create Dataset")

    @property
    def ids(self):
        """
        Return dictionary of parameters across results
        """
        ids = sc.ddict(set)
        for res in self.results.values():
            for k, v in res.id.items():
                ids[k].add(v)
        return {k: sorted(v) for k, v in ids.items()}

    def __repr__(self):
        s = "<Dataset:\n"
        for k, v in self.ids.items():
            s += f"\t'{k}':{v}\n"
        s += ">"
        return s

    def __getitem__(self, item):
        # Dict-like access for results
        return self.results[item]

    def __iter__(self):
        for result in self.results.values():
            yield result

    def __len__(self):
        return len(self.results)

    def __bool__(self):
        return len(self.results) > 0

    def filter(self, **kwargs):
        # Return results matching particular ids
        # kwargs: key-value pairs, that should be present in Sample.id, to filter results on
        # Can specify a list/set of values to match multiple items
        results = sc.odict()
        for res in self:
            for k, v in kwargs.items():
                if isinstance(v, list) or isinstance(v, set):
                    if res.id[k] not in v:
                        break
                elif res.id[k] != v:
                    break  # This result does not match, skip it
            else:
                results[res.identifier] = res

        return Dataset(results=results)

    def get(self, **kwargs):
        # Retrieve a single result from a filter operation
        # e.g. `res = Dataset.get(scenario='foo')` instead of
        # `res = Dataset.filter(scenario='foo')[0]` instead of
        # assuming that the arguments result in only 1 result being selected
        ds = self.filter(**kwargs)
        if len(ds) == 0:
            raise KeyError("No matching results were found")
        elif len(ds) > 1:
            raise KeyError("Multiple results matched the requested criteria")
        else:
            return ds[0]

identifiers_file = 'identifiers.txt'
summary_file = 'summary.csv'

class Samples:
    """
    Stores CSV outputs and summary dataframes

    To construct, use ``Samples.new()``. To read an existing one, use ``Samples(fname)``.
    The sample files are just ZIP archives with plain text CSV and TXT files so they
    can be easily accessed externally as well.
    """

    def __init__(self, fname, memory_buffer=True, preload=False):
        """

        Args:
            fname: Name of zip file to load
            memory_buffer: Load the file into memory. This avoids locking the file on disk and
                           improves performance when loading random parts of the file. This option can be
                           disabled in scripts to reduce memory requirements and improve performance if
                           not accessing the raw dataframes
            preload: Load all of the dataframes into memory during construction. This slows down initialization
                     but subsequent access to the dataframes will be fast - otherwise, there will be overhead incurred
                     the first time that the dataframes are used. Preloading is useful for interactive scripts where
                     it is known that the dataframes will be required. It should not be used if only the summary
                     is needed.
        """

        # Copy the file into a memory buffer - mainly so that we don't lock the file
        # on disk, but also this should theoretically improve performance when loading
        # random parts of the file.
        self._fname = fname
        self._zipfile = None

        if memory_buffer:
            with open(fname, "rb") as f:
                buffer = io.BytesIO(f.read())
            self._zipfile = zipfile.ZipFile(buffer, mode="r")

        # Read the identifiers
        with self.zipfile.open(identifiers_file) as f:
            identifiers = f.readlines()
            identifiers = [x.decode().strip() for x in identifiers]

        with self.zipfile.open(summary_file) as f:
            self.summary = sc.dataframe.read_csv(f)
            self.summary.set_index(identifiers, inplace=True)
            self.summary = self.summary.sort_index()

        self._cache = {}  # Cache the dataframes

        if preload:
            self.preload()

    def copy(self):
        """
        Shallow copy - shared cache, copied summary

        This allows efficient filtering of seeds within runs by removing rows from
        the copy's summary, while not reloading or duplicating any of the dataframes
        in memory
        """
        new = sc.cp(self)  # Shallow copy of
        new.summary = sc.dcp(self.summary)
        return new

    def preload(self):
        """
        Load all dataframes into cache

        This is done based on the seeds in self.seeds, therefore if some of the seeds
        are removed prior to preloading, then those dataframes will not be loaded
        """
        for seed in self.seeds:
            self[seed] # This automatically caches the loaded seeds
        return

    @property
    def zipfile(self):
        if self._zipfile:
            return self._zipfile
        else:
            return zipfile.ZipFile(self._fname, mode="r")

    def __repr__(self):
        return f"<Samples {'-'.join(str(x) for x in self.id.values())}, {len(self)} seeds>"

    @property
    def index(self):
        """Alias summary dataframe index"""
        return self.summary.index

    @property
    def columns(self):
        """Alias summary dataframe columns"""
        return self.summary.columns

    @property
    def identifier(self):
        """
        Return tuple identifier for this run

        The identifier is something like

            ('level_1',3,'level_5')

        Can be used to identify this result set in the context of sweeping over these identifiers
        e.g. the identifier could contain the starting level of restrictions. The dimensionality
        will vary depending on the analysis, hence it returns a tuple of arbitrary length. It would
        just be expected that all results being analyzed at the same time would have the same set of
        identifiers. The first two index levels are always 'beta' and 'seed', therefore these are
        dropped from the ID.
        """
        return tuple(self.summary.iloc[0].name[1:])

    @property
    def id(self):
        """
        Return a dictionary with the identifiers and associated values

        For example:

        >>> result.id
        (2.0, 'Gradually escalate restrictions', 0.16, 0.5, '95_70', 1)
        >>>  result.identifier
        {'beta_multiplier': 2.0,
         'strategy': 'Gradually escalate restrictions',
         'symp_test': 0.16,
         'vac_rel_test': 0.5,
         'vac_peak_coverage': '95_70',
         'incursions_per_day': 1}

        Returns: A dictionary {identifier name: value}
        """
        return dict(zip(tuple(self.summary.index.names[1:]), self.identifier))

    def __len__(self):
        """
        Return number of runs

        Returns: Integer number of seeds
        """
        return len(self.summary)

    def __iter__(self):
        """
        Iterate over dataframes

        For example:

        >>> for df in result:
        >>>     max(df['new_diagnoses'])

        Returns: Generator over dataframes
        """
        for seed in self.seeds:
            yield self[seed]

    def __contains__(self, seed):
        """ Check if a seed is contained """
        return seed in set(self.seeds)

    @property
    def seeds(self):
        """
        Return array of all seeds

        The seeds are 'registered' in the "seed" column of the summary dataframe.
        Therefore, to discard seeds, the rows can be dropped from the summary
        dataframe. In that case, iterating over `Samples.seeds()` will skip the excluded
        runs. Indexing into the Samples object will also fail to retrieve seeds that have
        been removed from the summary.

        Returns: Array of seed values
        """
        return self.index.get_level_values("seed").values
    
    @staticmethod
    def _seedfile(seed):
        """ Helper function to generate a seed file name """
        return f'seed_{seed}.csv'

    @classmethod
    def new(cls, folder, outputs, identifiers, fname=None):
        """
        Args:
            folder: The folder name
            outputs: A list of tuples (df:pd.DataFrame, summary_row:dict) where the summary row as an entry 'seed' for the seed
            identifiers: A list of columns to use as identifiers. These should appear in the summary dataframe and should have the same
                         value for all samples.
        """
        zipdata = {} # Store all the data to be written as files inside the zipfile
        summary_rows = []
        for df, row in outputs:
            zipdata[cls._seedfile(row["seed"])] = df.to_csv()
            summary_rows.append(row)

        # Write the identifier metadata
        if identifiers is None:
            identifiers = ["seed"]
        else:
            identifiers = ["seed"] + list(identifiers)
        identifiers = list(dict.fromkeys(identifiers))  # Unique, maintaining order
        zipdata[identifiers_file] = "\n".join(identifiers)

        # Write the summary
        summary = sc.dataframe(summary_rows)

        # Check that each identifier column only contains one unique value
        for identifier in identifiers:
            if summary[identifier].nunique() > 1 and identifier != 'seed':
                raise ValueError(f"Identifier column '{identifier}' contains more than one unique value")

        # Finish working on and write the summary
        summary.set_index(identifiers, inplace=True)
        zipdata[summary_file] = summary.to_csv()
        
        # Handle the zip file name
        if fname is None:
            fname = "-".join(str(row[x]) for x in identifiers[1:]) + ".zip"
        folder.mkdir(parents=True, exist_ok=True)
        
        # Save the zip file
        sc.savezip(folder/fname, data=zipdata, tobytes=False)

        return cls(folder/fname, memory_buffer=False)

    def get(self, seed):
        """
        Retrieve dataframe and summary row

        Use ``Samples[seed]`` to read only the dataframe.
        Use ``Samples.get(seed)`` to read both the dataframe and summary row
        """
        df = self[seed]
        row = sc.objdict(self.summary.xs(seed, level="seed").reset_index().iloc[0])
        return row, df

    def items(self):
        """
        Iterate over seeds and dataframes

        Example usage

        >>> res = Samples(...)
        >>> for seed, (row, df) in res:
        >>>     ...

        Returns: Tuple with
            - seed
            - Samples.get(seed) i.e. a tuple with
                - the summary dataframe row for the requested seed
                - the corresponding CSV output for that run
        """
        for seed in self.seeds:
            yield seed, self.get(seed)
            
    def __getitem__(self, item):
        """
        Overload getitem for convenience

        This provides two usages depending on the item

        - If the item is a number, it is treated as a seed
        - Otherwise, it is treated as a column in the summary dataframe
        - If the item is a two element tuple, the first element is a seed and the second element is a
          column in the individual run dataframe

        For example

        >>> a = results[0]
        >>> _, a = results.get(0)

        >>> a = results['cum_diagnoses']
        >>> a = results.summary['cum_diagnoses']

        >>> a = results[0,'new_diagnoses']
        >>> a = results[0]['new_diagnoses']

        Naturally, this means that the seeds must be integers (but this is the only data type
        that would make sense with Numpy's random number generator anyway)
        """
        if isinstance(item, tuple):
            return self[item[0]][item[1]]

        if not sc.isnumber(item):
            if item not in self.columns:
                sc.suggest(item, self.columns, die=True)
            return self.summary[item]

        # Otherwise, assume the item is a seed
        if item not in self:
            raise Exception(f'This dataset does not contain item "{item}"')

        if item not in self._cache:
            with self.zipfile.open(self._seedfile(item)) as f:
                self._cache[item] = sc.dataframe.read_csv(f, index_col="t")

        return self._cache[item].copy()

    def apply(self, fcn, *args, **kwargs):
        """
        Apply/map function to every dataframe

        The function will be applied to every individual dataframe in the collection.

        Args:
            fcn: A function to apply. It should take in a dataframe
            *args: Additional arguments for ``fcn``
            **kwargs: Additional arguments for ``fcn``

        Returns: A list with the output of ``fcn``
        """
        return [fcn(x, *args, **kwargs) for x in self]



import os
import runpy
from setuptools import setup, find_packages

# Get version
cwd = os.path.abspath(os.path.dirname(__file__))
versionpath = os.path.join(cwd, 'starsim', 'version.py')
version = runpy.run_path(versionpath)['__version__']

# Get the documentation
with open(os.path.join(cwd, 'README.rst'), "r") as f:
    long_description = f.read()

CLASSIFIERS = [
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Development Status :: 5 - Production/Stable",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
]

setup(
    name="starsim",
    version=version,
    author="Robyn Stuart, Cliff Kerr, Romesh Abeysuriya, Paula Sanz-Leon, Jamie Cohen, and Daniel Klein on behalf of the Starsim Collective",
    description="Starsim",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords=["agent-based model", "simulation", "disease", "epidemiology"],
    platforms=["OS Independent"],
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[ # NB: remember to update __init__.py if these requirements change
        'numpy',
        'scipy',
        'pandas>=2.0.0',
        'sciris>=3.1.6',
        'matplotlib',
        'numba',
        'networkx',
    ],
)



"""
Define interventions (and analyzers)
"""

import starsim as ss
import sciris as sc
import numpy as np

__all__ = ['Analyzer', 'Intervention']


class Analyzer(ss.Module):
    """
    Base class for analyzers. Analyzers are used to provide more detailed information 
    about a simulation than is available by default -- for example, pulling states 
    out of sim.people on a particular timestep before they get updated on the next step.
    
    The key method of the analyzer is ``apply()``, which is called with the sim
    on each timestep.
    
    To retrieve a particular analyzer from a sim, use sim.get_analyzer().
    """

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)
    
    def initialize(self, sim):
        return super().initialize(sim)
    
    def apply(self, sim):
        pass

    def finalize(self, sim):
        return super().finalize(sim)


class Intervention(ss.Module):
    """
    Base class for interventions.
    
    The key method of the intervention is ``apply()``, which is called with the sim
    on each timestep.
    """

    def __init__(self, eligibility=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eligibility = eligibility
        return

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)
    
    def initialize(self, sim):
        return super().initialize(sim)
    
    def apply(self, sim, *args, **kwargs):
        raise NotImplementedError

    def finalize(self, sim):
        return super().finalize(sim)

    def _parse_product(self, product):
        """
        Parse the product input
        """
        if isinstance(product, ss.Product):  # No need to do anything
            self.product = product
        elif isinstance(product, str):
            self.product = self._parse_product_str(product)
        else:
            errormsg = f'Cannot understand {product} - please provide it as a Product.'
            raise ValueError(errormsg)
        return

    def _parse_product_str(self, product):
        raise NotImplementedError

    def check_eligibility(self, sim):
        """
        Return an array of indices of agents eligible for screening at time t
        """
        if self.eligibility is not None:
            is_eligible = self.eligibility(sim)
        else:
            is_eligible = sim.people.alive  # Probably not required
        return is_eligible


# %% Template classes for routine and campaign delivery
__all__ += ['RoutineDelivery', 'CampaignDelivery']

class RoutineDelivery(Intervention):
    """
    Base class for any intervention that uses routine delivery; handles interpolation of input years.
    """

    def __init__(self, years=None, start_year=None, end_year=None, prob=None, annual_prob=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.years = years
        self.start_year = start_year
        self.end_year = end_year
        self.prob = sc.promotetoarray(prob)
        self.annual_prob = annual_prob  # Determines whether the probability is annual or per timestep
        self.coverage_dist = ss.bernoulli(p=0)  # Placeholder - initialize delivery
        return

    def initialize(self, sim):

        # Validate inputs
        if (self.years is not None) and (self.start_year is not None or self.end_year is not None):
            errormsg = 'Provide either a list of years or a start year, not both.'
            raise ValueError(errormsg)

        # If start_year and end_year are not provided, figure them out from the provided years or the sim
        if self.years is None:
            if self.start_year is None: self.start_year = sim.pars['start']
            if self.end_year is None:   self.end_year = sim.pars['end']
        else:
            self.start_year = self.years[0]
            self.end_year = self.years[-1]

        # More validation
        if not(any(np.isclose(self.start_year, sim.yearvec)) and any(np.isclose(self.end_year, sim.yearvec))):
            errormsg = 'Years must be within simulation start and end dates.'
            raise ValueError(errormsg)

        # Adjustment to get the right end point
        adj_factor = int(1 / sim.dt) - 1 if sim.dt < 1 else 1

        # Determine the timepoints at which the intervention will be applied
        self.start_point = sc.findfirst(sim.yearvec, self.start_year)
        self.end_point   = sc.findfirst(sim.yearvec, self.end_year) + adj_factor
        self.years       = sc.inclusiverange(self.start_year, self.end_year)
        self.timepoints  = sc.inclusiverange(self.start_point, self.end_point)
        self.yearvec     = np.arange(self.start_year, self.end_year + adj_factor, sim.dt)

        # Get the probability input into a format compatible with timepoints
        if len(self.years) != len(self.prob):
            if len(self.prob) == 1:
                self.prob = np.array([self.prob[0]] * len(self.timepoints))
            else:
                errormsg = f'Length of years incompatible with length of probabilities: {len(self.years)} vs {len(self.prob)}'
                raise ValueError(errormsg)
        else:
            self.prob = sc.smoothinterp(self.yearvec, self.years, self.prob, smoothness=0)

        # Lastly, adjust the probability by the sim's timestep, if it's an annual probability
        if self.annual_prob: self.prob = 1 - (1 - self.prob) ** sim.dt

        return


class CampaignDelivery(Intervention):
    """
    Base class for any intervention that uses campaign delivery; handles interpolation of input years.
    """

    def __init__(self, years, interpolate=None, prob=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.years = sc.promotetoarray(years)
        self.interpolate = True if interpolate is None else interpolate
        self.prob = sc.promotetoarray(prob)
        return

    def initialize(self, sim):
        # Decide whether to apply the intervention at every timepoint throughout the year, or just once.
        self.timepoints = sc.findnearest(sim.yearvec, self.years)

        if len(self.prob) == 1:
            self.prob = np.array([self.prob[0]] * len(self.timepoints))

        if len(self.prob) != len(self.years):
            errormsg = f'Length of years incompatible with length of probabilities: {len(self.years)} vs {len(self.prob)}'
            raise ValueError(errormsg)

        return


# %% Screening and triage
__all__ += ['BaseTest', 'BaseScreening', 'routine_screening', 'campaign_screening', 'BaseTriage', 'routine_triage',
            'campaign_triage']


class BaseTest(Intervention):
    """
    Base class for screening and triage.

    Args:
         product        (Product)       : the diagnostic to use
         prob           (float/arr)     : annual probability of eligible people receiving the diagnostic
         eligibility    (inds/callable) : indices OR callable that returns inds
         kwargs         (dict)          : passed to Intervention()
    """

    def __init__(self, product=None, prob=None, eligibility=None, **kwargs):
        super().__init__(**kwargs)
        self.prob = sc.promotetoarray(prob)
        self.eligibility = eligibility
        self._parse_product(product)
        self.screened = ss.State('screened', bool, False)
        self.screens = ss.State('screens', int, 0)
        self.ti_screened = ss.State('ti_screened', int, ss.INT_NAN)
        return

    def initialize(self, sim):
        Intervention.initialize(self, sim)
        self.outcomes = {k: np.array([], dtype=int) for k in self.product.hierarchy}
        return

    def deliver(self, sim):
        """
        Deliver the diagnostics by finding who's eligible, finding who accepts, and applying the product.
        """
        ti = sc.findinds(self.timepoints, sim.ti)[0]
        prob = self.prob[ti]  # Get the proportion of people who will be tested this timestep
        is_eligible = self.check_eligibility(sim)  # Check eligibility
        self.coverage_dist.set(p=prob)
        accept_uids = self.coverage_dist.filter(ss.true(is_eligible))
        if len(accept_uids):
            self.outcomes = self.product.administer(sim, accept_uids)  # Actually administer the diagnostic
        return accept_uids

    def check_eligibility(self, sim):
        raise NotImplementedError


class BaseScreening(BaseTest):
    """
    Base class for screening.
    Args:
        kwargs (dict): passed to BaseTest
    """
    def check_eligibility(self, sim):
        """
        Check eligibility
        """
        raise NotImplementedError

    def apply(self, sim, module=None):
        """
        Perform screening by finding who's eligible, finding who accepts, and applying the product.
        """
        accept_uids = np.array([])
        if sim.ti in self.timepoints:
            accept_uids = self.deliver(sim)
            self.screened[accept_uids] = True
            self.screens[accept_uids] += 1
            self.ti_screened[accept_uids] = sim.ti
            self.results['n_screened'][sim.ti] = len(accept_uids)
            self.results['n_dx'][sim.ti] = len(self.outcomes['positive'])

        return accept_uids


class BaseTriage(BaseTest):
    """
    Base class for triage.
    Args:
        kwargs (dict): passed to BaseTest
    """
    def check_eligibility(self, sim):
        return sc.promotetoarray(self.eligibility(sim))

    def apply(self, sim):
        self.outcomes = {k: np.array([], dtype=int) for k in self.product.hierarchy}
        accept_inds = np.array([])
        if sim.t in self.timepoints: accept_inds = self.deliver(sim)
        return accept_inds


class routine_screening(BaseScreening, RoutineDelivery):
    """
    Routine screening - an instance of base screening combined with routine delivery.
    See base classes for a description of input arguments.

    **Examples**::
        screen1 = ss.routine_screening(product=my_prod, prob=0.02) # Screen 2% of the eligible population every year
        screen2 = ss.routine_screening(product=my_prod, prob=0.02, start_year=2020) # Screen 2% every year starting in 2020
        screen3 = ss.routine_screening(product=my_prod, prob=np.linspace(0.005,0.025,5), years=np.arange(2020,2025)) # Scale up screening over 5 years starting in 2020
    """

    def __init__(self, product=None, prob=None, eligibility=None,
                 years=None, start_year=None, end_year=None, **kwargs):
        BaseScreening.__init__(self, product=product, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years)
        return

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim)  # Initialize this first, as it ensures that prob is interpolated properly
        BaseScreening.initialize(self, sim)  # Initialize this next
        return


class campaign_screening(BaseScreening, CampaignDelivery):
    """
    Campaign screening - an instance of base screening combined with campaign delivery.
    See base classes for a description of input arguments.

    **Examples**::

        screen1 = ss.campaign_screening(product=my_prod, prob=0.2, years=2030) # Screen 20% of the eligible population in 2020
        screen2 = ss.campaign_screening(product=my_prod, prob=0.02, years=[2025,2030]) # Screen 20% of the eligible population in 2025 and again in 2030
    """

    def __init__(self, product=None, sex=None, eligibility=None,
                 prob=None, years=None, interpolate=None, **kwargs):
        BaseScreening.__init__(self, product=product, sex=sex, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate)
        return

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim)
        BaseScreening.initialize(self, sim)  # Initialize this next
        return


class routine_triage(BaseTriage, RoutineDelivery):
    """
    Routine triage - an instance of base triage combined with routine delivery.
    See base classes for a description of input arguments.

    **Example**:
        # Example: Triage positive screens into confirmatory testing
        screened_pos = lambda sim: sim.get_intervention('screening').outcomes['positive']
        triage = ss.routine_triage(product=my_triage, eligibility=screen_pos, prob=0.9, start_year=2030)
    """

    def __init__(self, product=None, prob=None, eligibility=None,
                 years=None, start_year=None, end_year=None, annual_prob=None, **kwargs):
        BaseTriage.__init__(self, product=product, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years,
                                 annual_prob=annual_prob)
        return

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim)  # Initialize this first, as it ensures that prob is interpolated properly
        BaseTriage.initialize(self, sim)  # Initialize this next
        return


class campaign_triage(BaseTriage, CampaignDelivery):
    """
    Campaign triage - an instance of base triage combined with campaign delivery.
    See base classes for a description of input arguments.

    **Examples**:
        # Example: In 2030, triage all positive screens into confirmatory testing
        screened_pos = lambda sim: sim.get_intervention('screening').outcomes['positive']
        triage1 = hpv.campaign_triage(product=my_triage, eligibility=screen_pos, prob=0.9, years=2030)
    """

    def __init__(self, product=None, sex=None, eligibility=None,
                 prob=None, years=None, interpolate=None, annual_prob=None, **kwargs):
        BaseTriage.__init__(self, product=product, sex=sex, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate, annual_prob=annual_prob)
        return

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim)
        BaseTriage.initialize(self, sim)
        return


#%% Treatment interventions
__all__ += ['BaseTreatment', 'treat_num']


class BaseTreatment(Intervention):
    """
    Base treatment class.

    Args:
         product        (str/Product)   : the treatment product to use
         prob           (float/arr)     : probability of treatment aong those eligible
         eligibility    (inds/callable) : indices OR callable that returns inds
         kwargs         (dict)          : passed to Intervention()
    """
    def __init__(self, product=None, prob=None, eligibility=None, **kwargs):
        super().__init__(**kwargs)
        self.prob = sc.promotetoarray(prob)
        self.eligibility = eligibility
        self._parse_product(product)
        self.coverage_dist = ss.bernoulli(p=0)  # Placeholder
        return

    def initialize(self, sim):
        Intervention.initialize(self, sim)
        self.outcomes = {k: np.array([], dtype=int) for k in ['unsuccessful', 'successful']} # Store outcomes on each timestep
        return

    def get_accept_inds(self, sim):
        """
        Get indices of people who will acccept treatment; these people are then added to a queue or scheduled for receiving treatment
        """
        accept_uids = np.array([], dtype=int)
        eligible_uids = self.check_eligibility(sim)  # Apply eligiblity
        if len(eligible_uids):
            self.coverage_dist.set(p=self.prob[0])
            accept_uids = self.coverage_dist.filter(eligible_uids)
        return accept_uids

    def get_candidates(self, sim):
        """
        Get candidates for treatment on this timestep. Implemented by derived classes.
        """
        raise NotImplementedError

    def apply(self, sim):
        """
        Perform treatment by getting candidates, checking their eligibility, and then treating them.
        """
        # Get indices of who will get treated
        treat_candidates = self.get_candidates(sim)  # NB, this needs to be implemented by derived classes
        still_eligible = self.check_eligibility(sim)
        treat_uids = np.intersect1d(treat_candidates, still_eligible)
        if len(treat_uids):
            self.outcomes = self.product.administer(sim, treat_uids)
        return treat_uids


class treat_num(BaseTreatment):
    """
    Treat a fixed number of people each timestep.

    Args:
         max_capacity (int): maximum number who can be treated each timestep
    """
    def __init__(self, max_capacity=None, **kwargs):
        super().__init__(**kwargs)
        self.queue = []
        self.max_capacity = max_capacity
        return

    def add_to_queue(self, sim):
        """
        Add people who are willing to accept treatment to the queue
        """
        accept_inds = self.get_accept_inds(sim)
        if len(accept_inds): self.queue += accept_inds.tolist()
        return

    def get_candidates(self, sim):
        """
        Get the indices of people who are candidates for treatment
        """
        treat_candidates = np.array([], dtype=int)
        if len(self.queue):
            if self.max_capacity is None or (self.max_capacity > len(self.queue)):
                treat_candidates = self.queue[:]
            else:
                treat_candidates = self.queue[:self.max_capacity]
        return sc.promotetoarray(treat_candidates)

    def apply(self, sim):
        """
        Apply treatment. On each timestep, this method will add eligible people who are willing to accept treatment to a
        queue, and then will treat as many people in the queue as there is capacity for.
        """
        self.add_to_queue(sim)
        treat_inds = BaseTreatment.apply(self, sim) # Apply method from BaseTreatment class
        self.queue = [e for e in self.queue if e not in treat_inds] # Recreate the queue, removing people who were treated
        return treat_inds



#%% Vaccination


__all__ += ['BaseVaccination', 'routine_vx', 'campaign_vx']


class BaseVaccination(Intervention):
    """
    Base vaccination class for determining who will receive a vaccine.

    Args:
         product        (str/Product)   : the vaccine to use
         prob           (float/arr)     : annual probability of eligible population getting vaccinated
         eligibility    (inds/callable) : indices OR callable that returns inds
         label          (str)           : the name of vaccination strategy
         kwargs         (dict)          : passed to Intervention()
    """
    def __init__(self, product=None, prob=None, label=None, **kwargs):
        Intervention.__init__(self, **kwargs)
        self.prob = sc.promotetoarray(prob)
        self.label = label
        self._parse_product(product)
        self.vaccinated = ss.State('vaccinated', bool, False)
        self.n_doses = ss.State('doses', int, 0)
        self.ti_vaccinated = ss.State('ti_vaccinated', int, ss.INT_NAN)
        self.coverage_dist = ss.bernoulli(p=0)  # Placeholder
        return

    def apply(self, sim):
        """
        Deliver the diagnostics by finding who's eligible, finding who accepts, and applying the product.
        """
        accept_uids = np.array([])
        if sim.ti in self.timepoints:

            ti = sc.findinds(self.timepoints, sim.ti)[0]
            prob = self.prob[ti]  # Get the proportion of people who will be tested this timestep
            is_eligible = self.check_eligibility(sim)  # Check eligibility
            self.coverage_dist.set(p=prob)
            accept_uids = self.coverage_dist.filter(ss.true(is_eligible))

            if len(accept_uids):
                self.product.administer(sim.people, accept_uids)

                # Update people's state and dates
                self.vaccinated[accept_uids] = True
                self.ti_vaccinated[accept_uids] = sim.ti
                self.n_doses[accept_uids] += 1

        return accept_uids


class routine_vx(BaseVaccination, RoutineDelivery):
    """
    Routine vaccination - an instance of base vaccination combined with routine delivery.
    See base classes for a description of input arguments.
    """

    def __init__(self, product=None, prob=None, eligibility=None,
                 start_year=None, end_year=None, years=None, **kwargs):

        BaseVaccination.__init__(self, product=product, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years)
        return

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim)  # Initialize this first, as it ensures that prob is interpolated properly
        BaseVaccination.initialize(self, sim)  # Initialize this next
        return


class campaign_vx(BaseVaccination, CampaignDelivery):
    """
    Campaign vaccination - an instance of base vaccination combined with campaign delivery.
    See base classes for a description of input arguments.
    """

    def __init__(self, product=None, prob=None, eligibility=None,
                 years=None, interpolate=True, **kwargs):

        BaseVaccination.__init__(self, product=product, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate)
        return

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseVaccination.initialize(self, sim) # Initialize this next
        return





"""
Test the Dists object from distributions.py
"""

# %% Imports and settings
import numpy as np
import sciris as sc
import starsim as ss
import matplotlib.pyplot as pl

n = 5 # Default number of samples

def make_dist(name='test', **kwargs):
    """ Make a default Dist for testing """
    dist = ss.random(name=name, **kwargs)
    return dist

def make_dists(**kwargs):
    """ Make a Dists object with two distributions in it """
    distlist = [make_dist(), make_dist()]
    dists = ss.Dists(distlist)
    dists.initialize()
    return dists


# %% Define the tests

def test_seed():
    """ Test assignment of seeds """
    sc.heading('Testing assignment of seeds')
    
    # Create and initialize two distributions
    dists = make_dists()
    dist0, dist1 = dists.dists.values()

    print(f'Dists dist0 and dist1 were assigned seeds {dist0.seed} and {dist1.seed}, respectively')
    assert dist0.seed != dist1.seed
    return dist0, dist1


def test_reset(n=n):
    """ Sample, reset, sample """
    sc.heading('Testing sample, reset, sample')
    dists = make_dists()
    distlist = dists.dists.values()

    # Reset via the container, but only 
    before = sc.autolist()
    after = sc.autolist()
    for dist in distlist:
        before += list(dist(n))
    dists.reset() # Return to step 0
    for dist in distlist:
        after += list(dist(n))

    print(f'Initial sample:\n{before}')
    print(f'After reset sample:\n{after}')
    assert np.array_equal(before, after)
    
    return before, after


def test_jump(n=n):
    """ Sample, jump, sample """
    sc.heading('Testing sample, jump, sample')
    dists = make_dists()
    distlist = dists.dists.values()

    # Jump via the contianer
    before = sc.autolist()
    after = sc.autolist()
    for dist in distlist:
        before += list(dist(n))
    dists.jump(to=10) # Jump to 10th step
    for dist in distlist:
        after += list(dist(n))

    print(f'Initial sample:\n{before}')
    print(f'After jump sample:\n{after}')
    assert not np.array_equal(before, after)
    
    return before, after


def test_order(n=n):
    """ Ensure sampling from one RNG doesn't affect another """
    sc.heading('Testing from multiple random number generators to test if sampling order matters')
    dists = make_dists()
    d0, d1 = dists.dists.values()

    # Sample d0, d1
    before = d0(n)
    _ = d1(n)
    
    dists.reset()
    
    # Sample d1, d0
    _ = d1(n)
    after = d0(n)

    print(f'When sampling rng0 before rng1: {before}')
    print(f'When sampling rng0 after rng1: {after}')
    assert np.array_equal(before, after)

    return before, after


# %% Test a minimally different world

class CountInf(ss.Intervention):
    """ Store every infection state in a timepoints x people array """
    def initialize(self, sim):
        n_agents = len(sim.people)
        self.arr = np.zeros((sim.npts, n_agents))
        self.n_agents = n_agents
        return
    
    def apply(self, sim):
        self.arr[sim.ti, :] = np.array(sim.diseases.sir.infected)[:self.n_agents]
        return


class OneMore(ss.Intervention):
    """ Add one additional agent and infection """
    def apply(self, sim):
        if sim.ti == 0:
            # Create an extra agent
            preg = ss.Pregnancy(rel_fertility=0) # Ensure no default births
            preg.initialize(sim)
            new_uids = np.array([len(sim.people)]) # Hack since make_embryos doesn't return UIDs
            preg.make_embryos(sim, np.array([0])) # Assign 0th agent to be the "mother"
            assert len(new_uids) == 1
            sim.people.age[new_uids] = -100 # Set to a very low number to never reach debut age
            
            # Infect that agent
            sir = sim.diseases.sir
            sir.set_prognoses(sim, new_uids)
            sir.ti_recovered[new_uids] = sim.ti + 1 # Reset recovery time to next timestep
            
            # Reset the random states
            p = sir.pars
            for dist in [p.dur_inf, p.p_death]:
                dist.jump(sim.ti+1)

        return


def plot_infs(s1, s2):
    """ Compare infection arrays from two sims """
    a1 = s1.interventions.countinf.arr
    a2 = s2.interventions.countinf.arr
    
    fig = pl.figure()
    pl.subplot(1,3,1)
    pl.pcolormesh(a1.T)
    pl.xlabel('Timestep')
    pl.ylabel('Person')
    pl.title('Baseline')
    
    pl.subplot(1,3,2)
    pl.pcolormesh(a2.T)
    pl.title('OneMore')
    
    pl.subplot(1,3,3)
    pl.pcolormesh(a2.T - a1.T)
    pl.title('Difference')
    
    sc.figlayout()
    return fig


def test_worlds(do_plot=False):
    
    res = sc.objdict()
    
    pars = dict(
        start = 2000,
        end = 2100,
        n_agents = 200,
        verbose = 0.05,
        diseases = dict(
            type = 'sir',
            init_prev = 0.1,
            beta = 1.0,
            dur_inf = 20,
            p_death = 0, # Here since analyzer can't handle variable numbers of people
        ),
        networks = dict(
            type = 'embedding',
            duration = 5, # Must be shorter than dur_inf for SIR transmission to occur
        ),
    )
    s1 = ss.Sim(pars=pars, interventions=CountInf())
    s2 = ss.Sim(pars=pars, interventions=[CountInf(), OneMore()])
    
    s1.run()
    s2.run()
    
    sum1 = s1.summarize()
    sum2 = s2.summarize()
    res.sum1 = sum1
    res.sum2 = sum2
    
    if do_plot:
        s1.plot()
        plot_infs(s1, s2)
        pl.show()
    
    assert len(s2.people) == len(s1.people) + 1
    assert sum2.sir_cum_infections == sum1.sir_cum_infections + 1
    assert (s1.interventions.countinf.arr != s2.interventions.countinf.arr).sum() == 0
        
    return res


# %% Run as a script
if __name__ == '__main__':
    T = sc.timer()
    do_plot = True

    o1 = test_seed()
    o2 = test_reset(n)
    o3 = test_jump(n)
    o4 = test_order(n)
    o5 = test_worlds(do_plot=do_plot)

    T.toc()



"""
Define non-communicable disease (NCD) model
"""

import numpy as np
import starsim as ss
import sciris as sc

__all__ = ['NCD']

class NCD(ss.Disease):
    """
    Example non-communicable disease

    This class implements a basic NCD model with risk of developing a condition
    (e.g., hypertension, diabetes), a state for having the condition, and associated
    mortality.
    """
    def __init__(self, pars=None):
        default_pars = dict(
            initial_risk = ss.bernoulli(p=0.3), # Initial prevalence of risk factors
            #'affection_rate': ss.rate(p=0.1), # Instantaneous rate of acquisition applied to those at risk (units are acquisitions / year)
            dur_risk = ss.expon(scale=10),
            prognosis = ss.weibull(c=2, scale=5), # Time in years between first becoming affected and death
        )

        super().__init__(ss.omerge(default_pars, pars))
        self.add_states(
            ss.State('at_risk', bool, False),
            ss.State('affected', bool, False),
            ss.State('ti_affected', int, ss.INT_NAN),
            ss.State('ti_dead', int, ss.INT_NAN),
        )
        return

    @property
    def not_at_risk(self):
        return ~self.at_risk

    def set_initial_states(self, sim):
        """
        Set initial values for states. This could involve passing in a full set of initial conditions,
        or using init_prev, or other. Note that this is different to initialization of the State objects
        i.e., creating their dynamic array, linking them to a People instance. That should have already
        taken place by the time this method is called.
        """
        alive_uids = ss.true(sim.people.alive)
        initial_risk = self.pars['initial_risk'].filter(alive_uids)
        self.at_risk[initial_risk] = True
        self.ti_affected[initial_risk] = sim.ti + sc.randround(self.pars['dur_risk'].rvs(initial_risk) / sim.dt)
        return initial_risk

    def update_pre(self, sim):
        deaths = ss.true(self.ti_dead == sim.ti)
        sim.people.request_death(deaths)
        self.log.add_data(deaths, died=True)
        self.results.new_deaths[sim.ti] = len(deaths) # Log deaths attributable to this module
        return

    def make_new_cases(self, sim):
        new_cases = ss.true(self.ti_affected == sim.ti)
        self.affected[new_cases] = True
        prog_years = self.pars.prognosis.rvs(new_cases)
        self.ti_dead[new_cases] = sim.ti + sc.randround(prog_years / sim.dt)
        super().set_prognoses(sim, new_cases)
        return new_cases

    def init_results(self, sim):
        """
        Initialize results
        """
        super().init_results(sim)
        self.results += [
            ss.Result(self.name, 'n_not_at_risk', sim.npts, dtype=int),
            ss.Result(self.name, 'prevalence', sim.npts, dtype=float),
            ss.Result(self.name, 'new_deaths', sim.npts, dtype=int),
        ]
        return

    def update_results(self, sim):
        super().update_results(sim)
        ti = sim.ti
        alive = sim.people.alive
        self.results.n_not_at_risk[ti] = np.count_nonzero(self.not_at_risk & alive)
        self.results.prevalence[ti]    = np.count_nonzero(self.affected & alive)/alive.count()
        self.results.new_deaths[ti]    = np.count_nonzero(self.ti_dead == ti)
        return



"""
Define products
"""

import starsim as ss
import sciris as sc
import numpy as np


__all__ = ['Product', 'Dx', 'Tx', 'Vx']


class Product(ss.Module):
    """ Generic product implementation """

    def administer(self, people, inds):
        """ Adminster a Product - implemented by derived classes """
        raise NotImplementedError


class Dx(Product):
    """
    Generic class for diagnostics 
    """

    def __init__(self, df, hierarchy=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.df = df
        self.health_states = df.state.unique()
        self.diseases = df.disease.unique()

        if hierarchy is None:
            self.hierarchy = df.result.unique()  # Drawn from the order in which the outcomes are specified
        else:
            self.hierarchy = hierarchy

        # Create placehold for multinomial sampling
        n_results = len(self.hierarchy)
        self.result_dist = ss.choice(a=n_results)
        return

    @property
    def default_value(self):
        return len(self.hierarchy) - 1

    def administer(self, sim, uids, return_format='dict'):
        """
        Administer a testing product.
        
        Returns:

             if return_format=='array': an array of length len(inds) with integer entries that map each person to one of the result_states
             if return_format=='dict': a dictionary keyed by result_states with values containing the indices of people classified into this state
        """

        # Pre-fill with the default value, which is set to be the last value in the hierarchy
        results = sc.dataframe({'uids': uids, 'result': self.default_value})

        for disease in self.diseases:
            for state in self.health_states:
                this_state = getattr(sim.diseases[disease], state)
                these_uids = ss.true(this_state[uids])

                # Filter the dataframe to extract test results for people in this state
                df_filter = (self.df.state == state) & (self.df.disease == disease)
                thisdf = self.df[df_filter]  # apply filter to get the results for this state & genotype
                probs = [thisdf[thisdf.result == result].probability.values[0] for result in self.hierarchy]
                self.result_dist.pk = probs  # Overwrite distribution probabilities

                # Sort people into one of the possible result states and then update their overall results
                this_result = self.result_dist.rvs(these_uids)-these_uids # TODO: check!
                row_inds = results.uids.isin(these_uids)
                results.loc[row_inds, 'result'] = np.minimum(this_result, results.loc[row_inds, 'result'])

            if return_format == 'dict':
                output = {self.hierarchy[i]: results[results.result == i].uids.values for i in range(len(self.hierarchy))}
            elif return_format == 'array':
                output = results

        return output


class Tx(Product):
    """
    Treatment products change fundamental properties about People, including their prognoses and infectiousness.
    """
    def __init__(self, df, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.df = df
        self.diseases = df.disease.unique()
        self.health_states = df.state.unique()
        self.efficacy_dist = ss.bernoulli(p=0)
        return

    def administer(self, sim, uids, return_format='dict'):
        """
        Loop over treatment states to determine those who are successfully treated and clear infection
        """

        tx_successful = []  # Initialize list of successfully treated individuals

        for disease_name in self.diseases:

            disease = sim.diseases[disease_name]

            for state in self.health_states:

                pre_tx_state = getattr(disease, state)
                these_uids = ss.true(pre_tx_state[uids])

                if len(these_uids):

                    df_filter = (self.df.state == state) & (self.df.disease == disease_name)  # Filter by state
                    thisdf = self.df[df_filter]  # apply filter to get the results for this state & genotype

                    # Determine whether treatment is successful
                    self.efficacy_dist.set(p=thisdf.efficacy.values[0])

                    # HACK to reset the efficacy_dist as it is called multiple times per timestep. TODO: Refactor
                    self.efficacy_dist.jump(sim.ti+1)
                    eff_treat_inds = self.efficacy_dist.filter(these_uids)

                    post_tx_state_name = thisdf.post_state.values[0]
                    post_tx_state = getattr(disease, post_tx_state_name)

                    if len(eff_treat_inds):
                        tx_successful += list(eff_treat_inds)
                        pre_tx_state[eff_treat_inds] = False  # People who get treated effectively
                        post_tx_state[eff_treat_inds] = True

        tx_successful = np.array(list(set(tx_successful)))
        tx_unsuccessful = np.setdiff1d(uids, tx_successful)
        if return_format == 'dict':
            output = {'successful': tx_successful, 'unsuccessful': tx_unsuccessful}
        elif return_format == 'array':
            output = tx_successful

        return output


class Vx(Product):
    """ Vaccine product """
    def __init__(self, diseases=None, pars=None, par_dists=None, *args, **kwargs):
        pars = ss.omerge({}, pars)
        par_dists = ss.omerge({}, par_dists)
        super().__init__(pars, par_dists, *args, **kwargs)
        self.diseases = sc.tolist(diseases)

    def administer(self, people, uids):
        """ Apply the vaccine to the requested uids. """
        pass



"""
Define default gonorrhea disease module and related interventions
"""

import numpy as np
import starsim as ss

__all__ = ['Gonorrhea']

class Gonorrhea(ss.Infection):

    def __init__(self, pars=None, par_dists=None, *args, **kwargs):

        # States additional to the default disease states (see base class)
        # Additional states dependent on parameter values, e.g. self.p_symp?
        # These might be useful for connectors to target, e.g. if HIV reduces p_clear
        self.add_states(
            ss.State('symptomatic', bool, False),
            ss.State('ti_clearance', int, ss.INT_NAN),
            ss.State('p_symp', float, 1),
        )

        # Parameters
        pars = ss.omergeleft(pars,
            dur_inf_in_days = ss.lognorm_ex(mean=10, stdev=0.6),  # median of 10 days (IQR 7–15 days) https://sti.bmj.com/content/96/8/556
            p_symp = 0.5,  # Share of infections that are symptomatic. Placeholder value
            p_clear = 0.2,  # Share of infections that spontaneously clear: https://sti.bmj.com/content/96/8/556
            init_prev = 0.1,
        )

        par_dists = ss.omergeleft(par_dists,
            dur_inf_in_days = ss.lognorm_ex,
            p_symp          = ss.bernoulli,
            p_clear         = ss.bernoulli,
            init_prev       = ss.bernoulli,
        )

        super().__init__(pars=pars, par_dists=par_dists, *args, **kwargs)
        return

    def init_results(self, sim):
        """
        Initialize results
        """
        super().init_results(sim)
        self.results += ss.Result(self.name, 'new_clearances', sim.npts, dtype=int)
        return

    def update_results(self, sim):
        ti = sim.ti
        super(Gonorrhea, self).update_results(sim)
        self.results.n_symptomatic[ti] = self.symptomatic.count()
        self.results.new_clearances[ti] = np.count_nonzero(self.ti_clearance == ti)
        return

    def update_pre(self, sim):
        # What if something in here should depend on another module?
        # I guess we could just check for it e.g., 'if HIV in sim.modules' or
        # 'if 'hiv' in sim.people' or something
        # Natural clearance
        clearances = self.ti_clearance <= sim.ti
        self.susceptible[clearances] = True
        self.infected[clearances] = False
        self.symptomatic[clearances] = False
        self.ti_clearance[clearances] = sim.ti

        return
    
    def make_new_cases(self, sim):
        super(Gonorrhea, self).make_new_cases(sim)
        return

    def set_prognoses(self, sim, target_uids, source_uids=None):
        """
        Natural history of gonorrhea for adult infection
        """
        super().set_prognoses(sim, target_uids, source_uids)

        # Set infection status
        self.susceptible[target_uids] = False
        self.infected[target_uids] = True
        self.ti_infected[target_uids] = sim.ti

        # Set infection status
        symp_uids = self.pars.p_symp.filter(target_uids)
        self.symptomatic[symp_uids] = True

        # Set natural clearance
        clear_uids = self.pars.p_clear.filter(target_uids)
        dur = sim.ti + self.pars.dur_inf_in_days.rvs(clear_uids)/365/sim.dt # Convert from days to years and then adjust for dt
        self.ti_clearance[clear_uids] = dur

        return

    def set_congenital(self, sim, target_uids, source_uids=None):
        pass



"""
Result structures.
"""

import numpy as np
import sciris as sc
import starsim as ss


__all__ = ['Result', 'Results']


class Result(np.ndarray):
    
    def __new__(cls, module=None, name=None, shape=None, dtype=None, scale=None):
        arr = np.zeros(shape=shape, dtype=dtype).view(cls)
        arr.name = name
        arr.module = module
        arr.scale = scale
        return arr
    
    def __repr__(self):
        modulestr = f'{self.module}.' if (self.module is not None) else ''
        cls_name = self.__class__.__name__
        arrstr = super().__repr__().removeprefix(cls_name)
        out = f'{cls_name}({modulestr}{self.name}):\narray{arrstr}'
        return out

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.name = getattr(obj, 'name', None)
        self.module = getattr(obj, 'module', None)
        self.scale = getattr(obj, 'scale', None)
        return

    def __array_wrap__(self, obj, **kwargs):
        if obj.shape == ():
            return obj[()]
        else:
            return super().__array_wrap__(obj, **kwargs)
    
    def to_df(self):
        return sc.dataframe({self.name:self})
    

class Results(ss.ndict):
    
    def __init__(self, module, strict=True, *args, **kwargs):
        super().__init__(type=Result, strict=strict)
        if hasattr(module, 'name'):
            module = module.name
        self.setattribute('_module', module)
        return
    
    def append(self, arg, key=None):
        if isinstance(arg, (list, tuple)):
            result = ss.Result(self._module, *arg)
        elif isinstance(arg, dict):
            result = ss.Result(self._module, **arg)
        else:
            result = arg
        if result.module != self._module:
            result.module = self._module
        
        super().append(result, key=key)
        return
    
    def to_df(self):
        pass
    
    def __repr__(self, *args, **kwargs): # TODO: replace with dataframe summary
        return super().__repr__(*args, **kwargs)
        
    def plot(self):
        pass



from .version import __version__, __versiondate__, __license__
from .settings      import *
from .parameters    import *
from .utils         import *
from .distributions import *
from .states        import *
from .people        import *
from .modules       import *
from .network       import *
from .results       import *
from .demographics  import *
from .products      import *
from .interventions import *
from .demographics  import *
from .connectors    import *
from .disease       import *
from .diseases      import *
from .sim           import *
from .run           import *
from .samples       import *

# Assign the root folder
import sciris as sc
root = sc.thispath(__file__).parent

# Import the version and print the license
if options.verbose:
    print(__license__)

# Double-check key requirements -- should match setup.py
sc.require(['sciris>=3.1.6', 'pandas>=2.0.0', 'scipy', 'numba', 'networkx'], message=f'The following dependencies for Starsim {__version__} were not met: <MISSING>.')
del sc # Don't keep this in the module


"""
Define random-number-safe distributions.
"""

import numpy as np
import sciris as sc
import scipy.stats as sps
import starsim as ss
import matplotlib.pyplot as pl

__all__ = ['find_dists', 'dist_list', 'Dists', 'Dist']


def str2int(string, modulo=1_000_000):
    """
    Convert a string to an int
    
    Cannot use Python's built-in hash() since it's randomized for strings, but
    this is almost as fast (and 5x faster than hashlib).
    """
    return int.from_bytes(string.encode(), byteorder='big') % modulo


def find_dists(obj, verbose=False):
    """ Find all Dist objects in a parent object """
    out = sc.objdict()
    tree = sc.iterobj(obj, depthfirst=False, flatten=True)
    if verbose: print(f'Found {len(tree)} objects')
    for trace,val in tree.items():
        if isinstance(val, Dist):
            out[trace] = val
            if verbose: print(f'  {trace} is a dist ({len(out)})')
    return out


class Dists(sc.prettyobj):
    """ Class for managing a collection of Dist objects """

    def __init__(self, obj=None, *args, base_seed=None, sim=None):
        if len(args): obj = [obj] + list(args)
        self.obj = obj
        self.dists = None
        self.base_seed = base_seed
        self.sim = sim
        self.initialized = False
        if self.obj is not None:
            self.initialize()
        return

    def initialize(self, obj=None, base_seed=None, sim=None, force=False):
        """
        Set the base seed, find and initialize all distributions in an object
        
        In practice, the object is usually a Sim, but can be anything.
        """
        if base_seed:
            self.base_seed = base_seed
        sim = sim if (sim is not None) else self.sim
        obj = obj if (obj is not None) else self.obj
        if obj is None:
            errormsg = 'Must supply a container that contains one or more Dist objects, typically the sim'
            raise ValueError(errormsg)
        self.dists = find_dists(obj)
        for trace,dist in self.dists.items():
            if not dist.initialized or force:
                dist.initialize(trace=trace, seed=base_seed, sim=sim, force=force)
        self.check_seeds()
        self.initialized = True
        return self
    
    def check_seeds(self):
        """ Check that no two distributions share the same seed """
        checked = dict()
        for trace,dist in self.dists.items():
            seed = dist.seed
            if seed in checked.keys():
                raise DistSeedRepeatError(checked[seed], dist)
            else:
                checked[seed] = dist
        return

    def jump(self, to=None, delta=1):
        """ Advance all RNGs, e.g. to timestep "to", by jumping """
        out = sc.autolist()
        for dist in self.dists.values():
            out += dist.jump(to=to, delta=delta)
        return out

    def reset(self):
        """ Reset each RNG """
        out = sc.autolist()
        for dist in self.dists.values():
            out += dist.reset()
        return out


class Dist: # TODO: figure out why subclassing sc.prettyobj breaks isinstance
    """
    Base class for tracking one random number generator associated with one distribution,
    i.e. one decision per timestep.
    
    See ss.dist_list for a full list of supported distributions.
    
    Although it's possible in theory to define a custom distribution (i.e., not
    one from NumPy or SciPy), in practice this is difficult. The distribution needs
    to have both a way to return random variables (easy), as well as the probability
    point function (inverse CDF). In addition, the distribution must be able to
    take a NumPy RNG as its bit generator. It's easier to just use a default Dist
    (e.g., ss.random()), and then take its output as input (i.e., quantiles) for 
    whatever custom distribution you want to create.
    
    Args:
        dist (rv_generic): optional; a scipy.stats distribution (frozen or not) to get the ppf from
        distname (str): the name for this class of distribution (e.g. "uniform")
        name (str): the name for this particular distribution (e.g. "age_at_death")
        seed (int): the user-chosen random seed (e.g. 3)
        offset (int): the seed offset; will be automatically assigned (based on hashing the name) if None
        strict (bool): if True, require initialization and invalidate after each call to rvs()
        auto (bool): whether to auto-reset the state after each draw
        sim (Sim): usually determined on initialization; the sim to use as input to callable parameters
        module (Module): usually determined on initialization; the module to use as input to callable parameters
        kwargs (dict): parameters of the distribution
        
    **Examples**::
        
        dist = ss.Dist(sps.norm, loc=3)
        dist.rvs(10) # Return 10 normally distributed random numbers
    """
    def __init__(self, dist=None, distname=None, name=None, seed=None, offset=None, 
                 strict=True, auto=True, sim=None, module=None, debug=False, **kwargs):
        # If a string is provided as "dist" but there's no distname, swap the dist and the distname
        if isinstance(dist, str) and distname is None:
            distname = dist
            dist = None
        self.dist = dist # The type of distribution
        self.distname = distname
        self.name = name
        self.pars = sc.dictobj(kwargs) # The user-defined kwargs
        self.seed = seed # Usually determined once added to the container
        self.offset = offset
        self.module = module
        self.sim = sim
        self.slots = None # Created on initialization with a sim
        self.strict = strict
        self.auto = auto
        self.debug = debug
        
        # Auto-generated 
        self.rvs_func = None # The default function to call in make_rvs() to generate the random numbers
        self.dynamic_pars = None # Whether or not the distribution has array or callable parameters
        self._pars = None # Validated and transformed (if necessary) parameters
        self._n = None # Internal variable to keep track of "n" argument (usually size)
        self._size = None # Internal variable to keep track of actual number of random variates asked for
        self._uids = None # Internal variable to track currently-in-use UIDs
        self._slots = None # Internal variable to track currently-in-use slots
        
        # History and random state
        self.rng = None # The actual RNG generator for generating random numbers
        self.trace = None # The path of this object within the parent
        self.ind = 0 # The index of the RNG (usually updated on each timestep)
        self.called = 0 # The number of times the distribution has been called
        self.history = [] # Previous states
        self.ready = True
        self.initialized = False
        if not strict: # Otherwise, wait for a sim
            self.initialize()
        return
    
    def __repr__(self):
        """ Custom display to show state of object """
        tracestr = '<no trace>' if self.trace is None else f"{self.trace}"
        classname = self.__class__.__name__
        diststr = ''
        if classname == 'Dist':
            if self.dist is not None:
                try:
                    diststr = f'dist={self.dist.name}, '
                except:
                    try: # What is wrong with you, SciPy -- after initialization, it moves here
                        diststr = f'dist={self.dist.dist.name}, '
                    except:
                        diststr = f'dist={type(self.dist)}'
            elif self.distname is not None:
                diststr = f'dist={self.distname}, '
        string = f'ss.{classname}({tracestr}, {diststr}pars={dict(self.pars)})'
        return string
    
    def disp(self):
        """ Return full display of object """
        return sc.pr(self)
    
    def show_state(self, output=False):
        """ Show the state of the object """
        keys = ['pars', 'trace', 'offset', 'seed', 'ind', 'called', 'ready', 'state_int']
        data = {key:getattr(self,key) for key in keys}
        s = f'{self}'
        for key,val in data.items():
            s += f'\n{key:9s} = {val}'
        if output:
            return data
        else:
            print(s)
            return
    
    def __call__(self, n=1):
        """ Alias to self.rvs() """
        return self.rvs(n=n)
    
    def set(self, dist=None, **kwargs):
        """ Set (change) the distribution type, or one or more parameters of the distribution """
        if dist:
            self.dist = dist
            self.process_dist()
        if kwargs:
            self.pars.update(kwargs)
            self.process_pars(call=False)
        return

    @property
    def bitgen(self):
        try:    return self.rng.bit_generator
        except: return None
    
    @property
    def state(self):
        """ Get the current state """
        try:    return self.bitgen.state
        except: return None
        
    @property
    def state_int(self):
        """ Get the integer corresponding to the current state """
        try:    return self.state['state']['state']
        except: return None
    
    def get_state(self):
        """ Return a copy of the state """
        return self.state.copy()
    
    def make_history(self, reset=False):
        """ Store the current state in history """
        if reset:
            self.history = []
        self.history.append(self.get_state()) # Store the initial state
        return

    def reset(self, state=0):
        """
        Restore state, allowing the same numbers to be resampled
        
        Use 0 for original state, -1 for most recent state.
        
        **Example**::
            
            dist = ss.random(seed=5).initialize()
            r1 = dist(5)
            r2 = dist(5)
            dist.reset(-1)
            r3 = dist(5)
            dist.reset(0)
            r4 = dist(5)
            assert all(r1 != r2)
            assert all(r2 == r3)
            assert all(r4 == r1)
        """
        if not isinstance(state, dict):
            state = self.history[state]
        self.rng.bit_generator.state = state.copy()
        self.ready = True
        return self.state

    def jump(self, to=None, delta=1):
        """ Advance the RNG, e.g. to timestep "to", by jumping """
        jumps = to if (to is not None) else self.ind + delta
        self.ind = jumps
        self.reset() # First reset back to the initial state (used in case of different numbers of calls)
        if jumps: # Seems to randomize state if jumps=0
            self.bitgen.state = self.bitgen.jumped(jumps=jumps).state # Now take "jumps" number of jumps
        return self.state
    
    def initialize(self, trace=None, seed=None, module=None, sim=None, slots=None, force=False):
        """ Calculate the starting seed and create the RNG """
        
        if self.initialized is True and not force: # Don't warn if we have a partially initialized distribution
            msg = f'Distribution {self} is already initialized, use force=True if intentional'
            ss.warn(msg)
        
        # Calculate the offset (starting seed)
        self.process_seed(trace, seed)
        
        # Create the actual RNG
        self.rng = np.random.default_rng(seed=self.seed)
        self.make_history(reset=True)
        
        # Handle the sim, module, and slots
        self.sim = sim if (sim is not None) else self.sim
        self.module = module if (module is not None) else self.module
        if slots is None and self.sim is not None:
            try:
                slots = self.sim.people.slot
            except Exception as E:
                warnmsg = f'Could not extract slots from sim object, is this an error?\n{E}'
                ss.warn(warnmsg)
        if slots is not None:
            self.slots = slots
            
        # Initialize the distribution and finalize
        self.process_dist()
        self.process_pars(call=False)
        self.ready = True
        if self.trace is None or self.sim is None or self.slots is None:
            self.initialized = 'partial' # Initialized enough to produce random numbers, but not fully initialized
        else:
            self.initialized = True
        return self
    
    def process_seed(self, trace=None, seed=None):
        """ Obtain the seed offset by hashing the path to this distribution; called automatically """
        unique_name = trace or self.trace or self.name
        if unique_name:
            if not self.name:
                self.name = unique_name
            self.trace = unique_name
            self.offset = str2int(unique_name) # Key step: hash the path to the distribution
        else:
            self.offset = self.offset or 0
        self.seed = self.offset + (seed or self.seed or 0)
        return
    
    def process_dist(self):
        """ Ensure the distribution works """
        
        # Handle a SciPy distribution, if provided
        if self.dist is not None:
            
            # Pull out parameters of an already-frozen distribution
            if isinstance(self.dist, sps._distn_infrastructure.rv_frozen):
                if not self.initialized: # Don't do this more than once
                    self.pars = sc.dictobj(sc.mergedicts(self.pars, self.dist.kwds))
            
            # Convert to a frozen distribution
            if isinstance(self.dist, sps._distn_infrastructure.rv_generic):
                spars = self.process_pars(call=False)
                self.dist = self.dist(**spars) 
                
            # Override the default random state with the correct one
            self.dist.random_state = self.rng 
            
        # Set the default function for getting the rvs
        if self.distname is not None and hasattr(self.rng, self.distname): # Don't worry if it doesn't, it's probably being manually overridden
            self.rvs_func = self.distname # e.g. self.rng.uniform -- can't use the actual function because can become linked to the wrong RNG
        
        return
    
    def process_size(self, n=1):
        """ Handle an input of either size or UIDs and calculate size, UIDs, and slots """
        if np.isscalar(n) or isinstance(n, tuple):  # If passing a non-scalar size, interpret as dimension rather than UIDs iff a tuple
            uids = None
            slots = None
            size = n
        else:
            uids = np.asarray(n)
            if len(uids):
                slots = self.slots[uids]
                if len(slots): # Handle case where uids is boolean
                    size = slots.max() + 1
                else:
                    size = 0
            else:
                slots = np.array([])
                size = 0
        
        self._n = n
        self._size = size
        self._uids = uids
        self._slots = slots
        return size, slots
    
    def process_pars(self, call=True):
        """ Ensure the supplied dist and parameters are valid, and initialize them; called automatically """
        self._pars = sc.cp(self.pars) # The actual keywords; shallow copy, modified below for special cases
        if call:
            self.call_pars()
        spars = self.sync_pars()
        return spars
    
    def call_pars(self):
        """ Check if any parameters need to be called to be turned into arrays """
        
        # Initialize
        size, uids = self._size, self._uids
        if self.dynamic_pars != False: # Allow "False" to prevent ever using dynamic pars (used in ss.choice())
            self.dynamic_pars = None
        
        # Check each parameter
        for key,val in self._pars.items():
            
            # If the parameter is callable, then call it
            if callable(val): 
                size_par = uids if uids is not None else size
                out = val(self.module, self.sim, size_par) # TODO: swap order to sim, module, size?
                val = np.asarray(out) # Necessary since UIDArrays don't allow slicing # TODO: check if this is correct
                self._pars[key] = val
            
            # If it's iterable and UIDs are provided, then we need to use array-parameter logic
            if self.dynamic_pars is None and np.iterable(val) and uids is not None:
                self.dynamic_pars = True
        return
    
    def sync_pars(self):
        """ Perform any necessary synchronizations or transformations on distribution parameters """
        self.update_dist_pars()
        return self._pars
    
    def update_dist_pars(self, pars=None):
        """ Update SciPy distribution parameters """
        if self.dist is not None:
            pars = pars if pars is not None else self._pars
            self.dist.kwds = pars
        return
    
    def rand(self, size):
        """ Simple way to get simple random numbers """
        return self.rng.random(size)
    
    def make_rvs(self):
        """ Return default random numbers for scalar parameters """
        if self.rvs_func is not None:
            rvs_func = getattr(self.rng, self.rvs_func) # Can't store this because then it references the wrong RNG after copy
            rvs = rvs_func(**self._pars, size=self._size)
        elif self.dist is not None:
            rvs = self.dist.rvs(self._size)
        else:
            errormsg = 'Dist.rvs() failed: no valid NumPy/SciPy function found in this Dist. Has it been created and initialized correctly?'
            raise ValueError(errormsg)
        return rvs
    
    def ppf(self, rands):
        """ Return default random numbers for array parameters """
        rvs = self.dist.ppf(rands)
        return rvs
    
    def rvs(self, n=1, reset=False):
        """
        Get random variables
        
        Args:
            n (int/tuple/arr): if an int or tuple, return this many random variables; if an array, treat as UIDs
        """
        # Check for readiness
        if not self.initialized:
            raise DistNotInitializedError(self)
        if not self.ready and self.strict:
            raise DistNotReadyError(self)
        
        # Figure out size, UIDs, and slots
        size, slots = self.process_size(n)
        
        # Check if size is 0, then we can return
        if size == 0:
            return np.array([], dtype=int) # int dtype allows use as index, e.g. when filtering
        
        # Store the state
        self.make_history() # Store the pre-call state
        
        # Check if any keywords are callable -- parameters shouldn't need to be reprocessed otherwise
        if True: # self.dynamic_pars: # TODO: fix!!!!
            self.process_pars()
        
        # Actually get the random numbers
        if self.dynamic_pars:
            rands = self.rand(size)[slots] # Get random values 
            rvs = self.ppf(rands) # Convert to actual values via the PPF
        else:
            rvs = self.make_rvs() # Or, just get regular values
            if self._slots is not None:
                rvs = rvs[self._slots]
        
        # Tidy up
        self.called += 1
        if reset:
            self.reset(-1)
        elif self.auto: # TODO: check
            self.jump()
        elif self.strict:
            self.ready = False
        if self.debug:
            simstr = f'on ti={self.sim.ti} ' if self.sim else ''
            sizestr = f'with size={self._size}, '
            slotstr = f'Σ(slots)={self._slots.sum()}, ' if self._slots else '<no slots>, '
            rvstr = f'Σ(rvs)={rvs.sum():0.2f}, |rvs|={rvs.mean():0.4f}'
            pre_state = str(self.history[-1]['state']['state'])[-5:]
            post_state = str(self.state_int)[-5:]
            statestr = f"state {pre_state}→{post_state}"
            print(f'Debug: {self} called {simstr}{sizestr}{slotstr}{rvstr}, {statestr}')
            assert pre_state != post_state # Always an error if the state doesn't change after drawing random numbers
            
        return rvs


    def plot_hist(self, n=1000, bins=None, fig_kw=None, hist_kw=None):
        """ Plot the current state of the RNG as a histogram """
        pl.figure(**sc.mergedicts(fig_kw))
        rvs = self.rvs(n)
        self.reset(-1) # As if nothing ever happened
        pl.hist(rvs, bins=bins, **sc.mergedicts(hist_kw))
        pl.title(str(self))
        pl.xlabel('Value')
        pl.ylabel(f'Count ({n} total)')
        return rvs
        

#%% Specific distributions

# Add common distributions so they can be imported directly; assigned to a variable since used in help messages
dist_list = ['random', 'uniform', 'normal', 'lognorm_ex', 'lognorm_im', 'expon',
             'poisson', 'weibull', 'delta', 'randint', 'bernoulli', 'choice']
__all__ += dist_list


class random(Dist):
    """ Random distribution, values on interval (0,1) """
    def __init__(self, **kwargs):
        super().__init__(distname='random', **kwargs)
        return
    
    def ppf(self, rands):
        return rands


class uniform(Dist):
    """ Uniform distribution, values on interval (low, high) """
    def __init__(self, low=0.0, high=1.0, **kwargs):
        super().__init__(distname='uniform', low=low, high=high, **kwargs)
        return
    
    def ppf(self, rands):
        p = self._pars
        rvs = rands * (p.high - p.low) + p.low
        return rvs


class normal(Dist):
    """ Normal distribution, with mean=loc and stdev=scale """
    def __init__(self, loc=0.0, scale=1.0, **kwargs):
        super().__init__(distname='normal', dist=sps.norm, loc=loc, scale=scale, **kwargs)
        return


class lognorm_im(Dist):
    """
    Lognormal distribution, parameterized in terms of the "implicit" (normal)
    distribution, with mean=loc and stdev=scale (see lognorm_ex for comparison).
    
    Note: the "loc" parameter here does *not* correspond to the mean of the resulting
    random variates!
    
    **Example**::
        
        ss.lognorm_im(mean=2, sigma=1).rvs(1000).mean() # Should be roughly 10
    """
    def __init__(self, mean=0.0, sigma=1.0, **kwargs):
        super().__init__(distname='lognormal', dist=sps.lognorm, mean=mean, sigma=sigma, **kwargs)
        return
    
    def sync_pars(self):
        """ Translate between NumPy and SciPy parameters """
        p = self._pars
        spars = sc.dictobj()
        spars.s = p.sigma
        spars.scale = np.exp(p.mean)
        spars.loc = 0
        self.update_dist_pars(spars)
        return spars
    

class lognorm_ex(Dist):
    """
    Lognormal distribution, parameterized in terms of the "explicit" (lognormal)
    distribution, with mean=mean and stdev=stdev (see lognorm_im for comparison).
    
    **Example**::
        
        ss.lognorm_ex(mean=2, stdev=1).rvs(1000).mean() # Should be close to 2
    """
    def __init__(self, mean=1.0, stdev=1.0, **kwargs):
        super().__init__(distname='lognormal', dist=sps.lognorm, mean=mean, stdev=stdev, **kwargs)
        return
    
    def convert_ex_to_im(self):
        """
        Lognormal distributions can be specified in terms of the mean and standard
        deviation of the "explicit" lognormal distribution, or the "implicit" normal distribution.
        This function converts the parameters from the lognormal distribution to the
        parameters of the underlying (implicit) distribution, which are the form expected by NumPy's
        and SciPy's lognorm() distributions.
        """
        p = self._pars
        mean = p.pop('mean')
        stdev = p.pop('stdev')
        if mean <= 0:
            errormsg = f'Cannot create a lognorm_ex distribution with mean≤0 (mean={mean}); did you mean to use lognorm_im instead?'
            raise ValueError(errormsg)
        std2 = stdev**2
        mean2 = mean**2
        sigma_im = np.sqrt(np.log(std2/mean2 + 1)) # Computes stdev for the underlying normal distribution
        mean_im  = np.log(mean2 / np.sqrt(std2 + mean2)) # Computes the mean of the underlying normal distribution
        p.mean = mean_im
        p.sigma = sigma_im
        return mean_im, sigma_im
    
    def sync_pars(self):
        """ Convert from overlying to underlying parameters, then translate to SciPy """
        self.convert_ex_to_im()
        spars = lognorm_im.sync_pars(self) # Borrow sync_pars from lognorm_im
        return spars
    

class expon(Dist):
    """ Exponential distribution """
    def __init__(self, scale=1.0, **kwargs):
        super().__init__(distname='exponential', dist=sps.expon, scale=scale, **kwargs)
        return


class poisson(Dist):
    """ Exponential distribution """
    def __init__(self, lam=1.0, **kwargs):
        super().__init__(distname='poisson', dist=sps.poisson, lam=lam, **kwargs)
        return
    
    def sync_pars(self):
        """ Translate between NumPy and SciPy parameters """
        spars = dict(mu=self._pars.lam)
        self.update_dist_pars(spars)
        return spars


class randint(Dist):
    """ Random integers, values on the interval [low, high-1] (i.e. "high" is excluded) """
    def __init__(self, low=0, high=2,  **kwargs):
        super().__init__(distname='integers', dist=sps.randint, low=low, high=high, **kwargs)
        return
    

class weibull(Dist):
    """ Weibull distribution -- NB, uses SciPy rather than NumPy """
    def __init__(self, c=1, loc=0, scale=1,  **kwargs):
        super().__init__(distname='weibull', dist=sps.weibull_min, c=c, loc=loc, scale=scale, **kwargs)
        return
    
    def make_rvs(self):
        """ Use SciPy rather than NumPy to include the scale parameter """
        rvs = self.dist.rvs(self._size)
        return rvs


class delta(Dist):
    """ Delta distribution: equivalent to np.full() """
    def __init__(self, v=0, **kwargs):
        super().__init__(distname='delta', v=v, **kwargs)
        return
    
    def make_rvs(self):
        return np.full(self._size, self._pars.v)
    
    def ppf(self, rands): # NB: don't actually need to use random numbers here, but not worth the complexity of avoiding this
        return np.full(rands.shape, self._pars.v)


class bernoulli(Dist):
    """
    Bernoulli distribution: return True or False with the specified probability (which can be an array)
    
    Unlike other distributions, Bernoulli distributions have a filter() method,
    which returns elements of the array that return True.
    """
    def __init__(self, p=0.5, **kwargs):
        super().__init__(distname='bernoulli', p=p, **kwargs)
        return
    
    def make_rvs(self):
        rvs = self.rng.random(self._size) < self._pars.p # 3x faster than using rng.binomial(1, p, size)
        return rvs
    
    def ppf(self, rands):
        rvs = rands < self._pars.p
        return rvs
    
    def filter(self, uids, both=False):
        """ Return UIDs that correspond to True, or optionally return both True and False """
        bools = self.rvs(uids)
        if both:
            return uids[bools], uids[~bools]
        else:
            return uids[bools]


class choice(Dist):
    """
    Random choice between discrete options
    
    **Examples**::
        
        # Simulate 10 die rolls
        ss.choice(6)(10) + 1 
        
        # Choose between specified options each with a specified probability (must sum to 1)
        ss.choice(a=[30, 70], p=[0.3, 0.7])(10)
    """
    def __init__(self, a=2, p=None, **kwargs):
        super().__init__(distname='choice', a=a, p=p, **kwargs)
        self.dynamic_pars = False # Set to false since array arguments don't imply dynamic pars here
        return
    
    def ppf(self, rands):
        """ Shouldn't actually be needed since dynamic pars not supported """
        pars = self._pars
        if np.isscalar(pars.a):
            pars.a = np.arange(pars.a)
        pcum = np.cumsum(pars.p)
        inds = np.searchsorted(pcum, rands)
        rvs = pars.a[inds]
        return rvs


#%% Dist exceptions

class DistNotInitializedError(RuntimeError):
    """ Raised when Dist object is called when not initialized. """
    def __init__(self, dist):
        msg = f'{dist} has not been initialized; please call dist.initialize()'
        super().__init__(msg)

class DistNotReadyError(RuntimeError):
    """ Raised when a Dist object is called without being ready. """
    def __init__(self, dist):
        msg = f'{dist} is not ready. This is likely caused by calling a distribution multiple times in a single step. Call dist.jump() to reset.'
        super().__init__(msg)
        
class DistSeedRepeatError(RuntimeError):
    """ Raised when a Dist object shares a seed with another """
    def __init__(self, dist1, dist2):
        msg = f'A common seed was found between {dist1} and {dist2}. This is likely caused by incorrect initialization of the parent Dists object.'
        super().__init__(msg)
